"""
This package contains the public API for eFLIPS-Depot.

It is to be used in conjunction with the
`eflips.model <https://github.com/mpm-tu-berlin/eflips-model>`_ package, where the Scenario is defined.

Notes on the usage of the API
-----------------------------

The following steps are recommended for using the API:

1. Check if there are already "driving events" in the database. They come from a "consumption simulation" and are
   associated with a vehicle. If there are no driving events, you may use the :func:`simple_consumption_simulation`
   (with ``initialize_vehicles=True``) to create them. This function will also initialize the vehicles in the database
   with the correct vehicle type and assign them to rotations.
2. Check if there is already a depot layout in the database. If there is not, you may use the
   :func:`generate_depot_layout` function to create a simple depot layout and plan.
3. Either use the :func:`simulate_scenario` function to run the whole simulation in one go, or use the following steps:
    a. Use the :func:`init_simulation` function to create a simulation host, which is a "black box" object containing
       all input data for the simulation.
    b. Use the :func:`run_simulation` function to run the simulation and obtain the results.
    c. Use the :func:`add_evaluation_to_database` function to add the results to the database.
4. For the results to be valid, the consumption simulation should now be run again.
    a. If you are using an external consumption model, run it again making sure it does not create new vehicles.
    b. Run the :func:`simple_consumption_simulation` function again, this time with ``initialize_vehicles=False``.
"""
import copy
import datetime
import logging
import os
import warnings
from collections import OrderedDict
from datetime import timedelta
from enum import Enum
from math import ceil
from typing import Any, Dict, Optional, Union

import sqlalchemy.orm
from eflips.model import (
    Area,
    Depot,
    Event,
    EventType,
    Rotation,
    Scenario,
    Trip,
    Vehicle,
    VehicleType,
    AreaType,
)
from sqlalchemy.orm import Session

import eflips.depot
from eflips.depot import (
    DepotEvaluation,
    SimulationHost,
)
from eflips.depot.api.private.depot import (
    delete_depots,
    depot_to_template,
    group_rotations_by_start_end_stop,
    generate_depot,
)
from eflips.depot.api.private.results_to_database import (
    get_finished_schedules_per_vehicle,
    generate_vehicle_events,
    complete_standby_departure_events,
    add_soc_to_events,
    add_events_into_database,
    update_vehicle_in_rotation,
    update_waiting_events,
)
from eflips.depot.api.private.smart_charging import optimize_charging_events_even
from eflips.depot.api.private.util import (
    create_session,
    repeat_vehicle_schedules,
    start_and_end_times,
    vehicle_type_to_global_constants_dict,
    VehicleSchedule,
    check_depot_validity,
)


class SmartChargingStrategy(Enum):
    """Enum class for different smart charging strategies."""

    NONE = 0
    """
    Do not use smart charging.

    Buses are charged with the maximum power available, from the time they arrive at the depot
    until they are full (or leave the depot).
    """
    EVEN = 1
    """
    Use smart charging with an even distribution of charging power over the time the bus is at the depot.

    This aims to
    minimize the peak power demand.
    """
    MIN_PRICE = 2
    """
    Use smart charging in order to minimize the cost of charging.

    The price profile can be specified using the
    PRICE_PROFILE environment variable. If this is not set, the price is loaded using an API.
    """


def simple_consumption_simulation(
    scenario: Union[Scenario, int, Any],
    initialize_vehicles: bool,
    database_url: Optional[str] = None,
    calculate_timeseries: bool = False,
    terminus_deadtime: timedelta = timedelta(minutes=1),
) -> None:
    """
    A simple consumption simulation and vehicle initialization.

    Energy consumotion is calculated by multiplying the vehicle's total distance by a constant
    ``VehicleType.consumption``.

    If run with ``initialize_vehicles=True``, the method will also initialize the vehicles in the database with the
    correct vehicle type and assign them to rotations. If this is false, it will assume that there are already vehicle
    entries and ``Rotation.vehicle_id`` is already set.

    :param scenario: Either a :class:`eflips.model.Scenario` object containing the input data for the simulation. Or
        an integer specifying the ID of a scenario in the database. Or any other object that has an attribute
        ``id`` that is an integer. If no :class:`eflips.model.Scenario` object is passed, the ``database_url``
        parameter must be set to a valid database URL ot the environment variable ``DATABASE_URL`` must be set to a
        valid database URL.

    :param initialize_vehicles: A boolean flag indicating whether the vehicles should be initialized in the database.
        When running this function for the first time, this should be set to True. When running this function again
        after the vehicles have been initialized, this should be set to False.

    :param database_url: An optional database URL. If no database URL is passed and the `scenario` parameter is not a
        :class:`eflips.model.Scenario` object, the environment variable `DATABASE_URL` must be set to a
        valid database URL.

    :param calculate_timeseries: A boolean flag indicating whether the timeseries should be calculated. If this is set
        to True, the SoC at each stop is calculated and added to the "timeseries" column of the Event table. If this
        is set to False, the "timeseries" column of the Event table will be set to ``None``. Setting this to false
        may significantly speed up the simulation.

    :param terminus_deadtime: The total deadtime taken to both attach and detach the charging cable at the terminus.
                              If the total deadtime is greater than the time between the arrival and departure of the
                              vehicle at the terminus, the vehicle will not be able to charge at the terminus.

    :return: Nothing. The results are added to the database.
    """
    logger = logging.getLogger(__name__)

    with create_session(scenario, database_url) as (session, scenario):
        rotations = (
            session.query(Rotation)
            .filter(Rotation.scenario_id == scenario.id)
            .order_by(Rotation.id)
            .options(sqlalchemy.orm.joinedload(Rotation.trips).joinedload(Trip.route))
            .options(sqlalchemy.orm.joinedload(Rotation.vehicle_type))
            .options(sqlalchemy.orm.joinedload(Rotation.vehicle))
        )
        if initialize_vehicles:
            for rotation in rotations:
                vehicle = Vehicle(
                    vehicle_type_id=rotation.vehicle_type_id,
                    scenario_id=scenario.id,
                    name=f"Vehicle for rotation {rotation.id}",
                )
                session.add(vehicle)
                rotation.vehicle = vehicle

                # Additionally, add a short STANDBY event with 100% SoC immediately before the first trip
                first_trip_start = rotation.trips[0].departure_time
                standby_start = first_trip_start - timedelta(seconds=1)
                standby_event = Event(
                    scenario_id=scenario.id,
                    vehicle_type_id=rotation.vehicle_type_id,
                    vehicle=vehicle,
                    station_id=rotation.trips[0].route.departure_station_id,
                    subloc_no=0,
                    time_start=standby_start,
                    time_end=first_trip_start,
                    soc_start=1,
                    soc_end=1,
                    event_type=EventType.CHARGING_OPPORTUNITY,
                    description=f"DUMMY Initial standby event for rotation {rotation.id}.",
                    timeseries=None,
                )
                session.add(standby_event)
        else:
            for rotation in rotations:
                if rotation.vehicle is None:
                    raise ValueError(
                        "The rotation does not have a vehicle assigned to it."
                    )

            vehicles = (
                session.query(Vehicle).filter(Vehicle.scenario_id == scenario.id).all()
            )
            for vehicle in vehicles:
                if (
                    session.query(Event).filter(Event.vehicle_id == vehicle.id).count()
                    == 0
                ):
                    # Also add a dummy standby-departure event if this vehicle has no events
                    rotation_per_vehicle = sorted(
                        vehicle.rotations, key=lambda r: r.trips[0].departure_time
                    )
                    earliest_trip = rotation_per_vehicle[0].trips[0]
                    area = (
                        session.query(Area)
                        .filter(Area.scenario_id == scenario.id)
                        .filter(Area.vehicle_type_id == Vehicle.vehicle_type_id)
                        .first()
                    )

                    standby_start = earliest_trip.departure_time - timedelta(seconds=1)
                    standby_event = Event(
                        scenario_id=scenario.id,
                        vehicle_type_id=vehicle.vehicle_type_id,
                        vehicle=vehicle,
                        station_id=area.depot.station_id,
                        area_id=area.id,
                        subloc_no=area.capacity,
                        time_start=standby_start,
                        time_end=earliest_trip.departure_time,
                        soc_start=1,
                        soc_end=1,
                        event_type=EventType.STANDBY_DEPARTURE,
                        description=f"DUMMY Initial standby event for rotation {earliest_trip.rotation_id}.",
                        timeseries=None,
                    )
                    session.add(standby_event)

        # Since we are doing no_autoflush blocks later, we need to flush the session once here so that unflushed stuff
        # From preceding functions is visible in the database
        session.flush()

        for rotation in rotations:
            rotation: Rotation
            with session.no_autoflush:
                vehicle_type = (
                    session.query(VehicleType)
                    .join(Rotation)
                    .filter(Rotation.id == rotation.id)
                    .one()
                )
                vehicle = (
                    session.query(Vehicle)
                    .join(Rotation)
                    .filter(Rotation.id == rotation.id)
                    .one()
                )
            if vehicle_type.consumption is None:
                raise ValueError(
                    "The vehicle type does not have a consumption value set."
                )
            consumption = vehicle_type.consumption

            # The departure SoC for this rotation is the SoC of the last event preceding the first trip
            with session.no_autoflush:
                current_soc = (
                    session.query(Event.soc_end)
                    .filter(Event.vehicle_id == rotation.vehicle_id)
                    .filter(Event.time_end <= rotation.trips[0].departure_time)
                    .order_by(Event.time_end.desc())
                    .first()[0]
                )

            for trip in rotation.trips:
                # Set up a timeseries
                soc_start = current_soc
                if calculate_timeseries and len(trip.stop_times) > 0:
                    timeseries = {
                        "time": [],
                        "soc": [],
                        "distance": [],
                    }
                    for i in range(len(trip.stop_times)):
                        current_time = trip.stop_times[i].arrival_time
                        dwell_duration = trip.stop_times[i].dwell_duration
                        elapsed_distance = trip.route.assoc_route_stations[
                            i
                        ].elapsed_distance
                        elapsed_energy = consumption * (elapsed_distance / 1000)  # kWh
                        soc = (
                            current_soc - elapsed_energy / vehicle_type.battery_capacity
                        )
                        timeseries["time"].append(current_time.isoformat())
                        timeseries["soc"].append(soc)
                        timeseries["distance"].append(elapsed_distance)
                        if dwell_duration > timedelta(seconds=0):
                            timeseries["time"].append(
                                (current_time + dwell_duration).isoformat()
                            )
                            timeseries["soc"].append(soc)
                            timeseries["distance"].append(elapsed_distance)
                else:
                    timeseries = None
                energy_used = consumption * trip.route.distance / 1000  # kWh
                current_soc = soc_start - energy_used / vehicle_type.battery_capacity

                # Create a driving event
                current_event = Event(
                    scenario_id=scenario.id,
                    vehicle_type_id=rotation.vehicle_type_id,
                    vehicle=vehicle,
                    trip_id=trip.id,
                    time_start=trip.departure_time,
                    time_end=trip.arrival_time,
                    soc_start=soc_start,
                    soc_end=current_soc,
                    event_type=EventType.DRIVING,
                    description=f"`VehicleType.consumption`-based driving event for trip {trip.id}.",
                    timeseries=timeseries,
                )
                session.add(current_event)

                # If the vehicle is
                #  - Capable of opportunity charging
                #  - On a Rotation which allows opportunity charging
                #  - Currently at a station which allows opportunity charging
                #  - which is not the last trip of the rotation
                #  We add a charging event

                if (
                    rotation.vehicle_type.opportunity_charging_capable
                    and rotation.allow_opportunity_charging
                    and trip.route.arrival_station.is_electrified
                    and trip != rotation.trips[-1]
                ):
                    logger.debug(
                        f"Adding opportunity charging event for trip {trip.id}"
                    )
                    # Identify the break time between trips
                    trip_index = rotation.trips.index(trip)
                    next_trip = rotation.trips[trip_index + 1]
                    break_time = next_trip.departure_time - trip.arrival_time

                    # How much energy can be charged in this time?
                    energy_charged = (
                        max([v[1] for v in vehicle_type.charging_curve])
                        * (
                            break_time.total_seconds()
                            - terminus_deadtime.total_seconds()
                        )
                        / 3600
                    )

                    if energy_charged > 0:
                        # Calculate the end SoC
                        post_charge_soc = min(
                            current_soc
                            + energy_charged / vehicle_type.battery_capacity,
                            1,
                        )

                        # Create a simple timeseries for the charging event
                        timeseries = {
                            "time": [
                                trip.arrival_time.isoformat(),
                                (trip.arrival_time + terminus_deadtime / 2).isoformat(),
                                (
                                    next_trip.departure_time - terminus_deadtime / 2
                                ).isoformat(),
                                next_trip.departure_time.isoformat(),
                            ],
                            "soc": [
                                current_soc,
                                current_soc,
                                post_charge_soc,
                                post_charge_soc,
                            ],
                        }

                        # Create the charging event
                        current_event = Event(
                            scenario_id=scenario.id,
                            vehicle_type_id=rotation.vehicle_type_id,
                            vehicle=vehicle,
                            station_id=trip.route.arrival_station_id,
                            time_start=trip.arrival_time,
                            time_end=next_trip.departure_time,
                            soc_start=current_soc,
                            soc_end=post_charge_soc,
                            event_type=EventType.CHARGING_OPPORTUNITY,
                            description=f"Opportunity charging event for trip {trip.id}.",
                            timeseries=timeseries,
                        )
                        current_soc = post_charge_soc
                        session.add(current_event)


def generate_depot_layout(
    scenario: Union[Scenario, int, Any],
    charging_power: float = 90,
    database_url: Optional[str] = None,
    delete_existing_depot: bool = False,
) -> None:
    """
    Generates one or more depots for the scenario.

    First, the rotations are scanned to identify all the spots that serve as start *and* end of a rotation. Then the set
    of rotations for these spots are checked for vehicle types that are used there. Next, the amount of vehicles that
    are simultaneously present at the depot is calculated. Then a depot layout with an arrival and a charging area for
    each vehicle type is created. The capacity of each area is taken from the calculated amount of vehicles.
    The depot layout is then added to the database.

    A default plan will also be generated, which includes the following default processes: standby_arrival, cleaning,
    charging and standby_departure. Each vehicle will be processed with this exact order (standby_arrival is optional
    because it only happens if a vehicle needs to wait for the next process).

    The function only deletes the depot if the `delete_existing_depot` parameter is set to True. If there is already a
    depot existing in this scenario and this parameter is set to False, a ValueError will be raised.

    :param scenario: Either a :class:`eflips.model.Scenario` object containing the input data for the simulation. Or
        an integer specifying the ID of a scenario in the database. Or any other object that has an attribute
        ``id`` that is an integer. If no :class:`eflips.model.Scenario` object is passed, the ``database_url``
        parameter must be set to a valid database URL ot the environment variable ``DATABASE_URL`` must be set to a
        valid database URL.

    :param charging_power: the charging power of the charging area in kW

    :param delete_existing_depot: if there is already a depot existing in this scenario, set True to delete this
        existing depot. Set to False and a ValueError will be raised if there is a depot in this scenario.

    :return: None. The depot layout will be added to the database.
    """
    with create_session(scenario, database_url) as (session, scenario):
        # Handles existing depot
        if session.query(Depot).filter(Depot.scenario_id == scenario.id).count() != 0:
            if delete_existing_depot is False:
                raise ValueError("Depot already exists.")
            delete_depots(scenario, session)

        # Identify all the spots that serve as start *and* end of a rotation
        for (
            first_last_stop_tup,
            vehicle_type_dict,
        ) in group_rotations_by_start_end_stop(scenario.id, session).items():
            first_stop, last_stop = first_last_stop_tup
            if first_stop != last_stop:
                raise ValueError("First and last stop of a rotation are not the same.")

            # Create one direct slot for each rotation (it's way too much, but should work)
            vt_capacity_dict: Dict[VehicleType, Dict[AreaType, None | int]] = {}
            for vehicle_type, rotations in vehicle_type_dict.items():
                rotation_count = len(rotations)
                vt_capacity_dict[vehicle_type] = {
                    AreaType.LINE: None,
                    AreaType.DIRECT_ONESIDE: rotation_count,
                    AreaType.DIRECT_TWOSIDE: None,
                }
            generate_depot(
                vt_capacity_dict,
                first_stop,
                scenario,
                session,
                charging_power=charging_power,
                num_shunting_slots=max(rotation_count // 10, 1),
                num_cleaning_slots=max(rotation_count // 10, 1),
            )


def apply_even_smart_charging(
    scenario: Union[Scenario, int, Any],
    database_url: Optional[str] = None,
    standby_departure_duration: timedelta = timedelta(minutes=5),
) -> None:
    """
    Takes a scenario where depot simulation has been run and applies smart charging to the depot.

    This modifies the time and power of the charging events in the database. The arrival and departure times and SoCs at
    these times are not modified.

    :param scenario: A :class:`eflips.model.Scenario` object containing the input data for the simulation.

    :param database_url: An optional database URL. If no database URL is passed and the `scenario` parameter is not a
        :class:`eflips.model.Scenario` object, the environment variable `DATABASE_URL` must be set to a valid database
        URL.

    :param standby_departure_duration: The duration of the STANDBY_DEPARTURE event. This is the time the vehicle is
        allowed to wait at the depot before it has to leave. The default is 5 minutes.

    :return: None. The results are added to the database.
    """
    logger = logging.getLogger(__name__)

    with create_session(scenario, database_url) as (session, scenario):
        depots = session.query(Depot).filter(Depot.scenario_id == scenario.id).all()
        for depot in depots:
            # Load all the charging events at this depot
            charging_events = (
                session.query(Event)
                .join(Area)
                .filter(Area.depot_id == depot.id)
                .filter(Event.event_type == EventType.CHARGING_DEPOT)
                .all()
            )

            # For each event, take the subsequent STANDBY_DEPARTURE event of the same vehicle
            # Reduce the STANDBY_DEPARTURE events duration to 5 minutes
            # Move the end time of the charging event to the start time of the STANDBY_DEPARTURE event
            for charging_event in charging_events:
                next_event = (
                    session.query(Event)
                    .filter(Event.time_start >= charging_event.time_end)
                    .filter(Event.vehicle_id == charging_event.vehicle_id)
                    .order_by(Event.time_start)
                    .first()
                )

                if (
                    next_event is None
                    or next_event.event_type != EventType.STANDBY_DEPARTURE
                ):
                    logger.info(
                        f"Event {charging_event.id} has no STANDBY_DEPARTURE event after a CHARGING_DEPOT "
                        f"event. No room for smart charging."
                    )
                    continue

                assert next_event.time_start == charging_event.time_end

                if (
                    next_event.time_end - next_event.time_start
                ) > standby_departure_duration:
                    next_event.time_start = (
                        next_event.time_end - standby_departure_duration
                    )
                    session.flush()
                    # Add a timeseries to the charging event
                    assert charging_event.timeseries is None
                    charging_event.timeseries = {
                        "time": [
                            charging_event.time_start.isoformat(),
                            charging_event.time_end.isoformat(),
                            next_event.time_start.isoformat(),
                        ],
                        "soc": [
                            charging_event.soc_start,
                            charging_event.soc_end,
                            charging_event.soc_end,
                        ],
                    }
                    charging_event.time_end = next_event.time_start
                    session.flush()

            optimize_charging_events_even(charging_events)


def simulate_scenario(
    scenario: Union[Scenario, int, Any],
    repetition_period: Optional[timedelta] = None,
    database_url: Optional[str] = None,
    smart_charging_strategy: SmartChargingStrategy = SmartChargingStrategy.EVEN,
) -> None:
    """
    This method simulates a scenario and adds the results to the database.

    It fills in the "Charging Events" in the :class:`eflips.model.Event` table and associates
    :class:`eflips.model.Vehicle` objects with all the existing "Driving Events" in the :class:`eflips.model.Event`
    table.

    :param scenario: Either a :class:`eflips.model.Scenario` object containing the input data for the simulation. Or
        an integer specifying the ID of a scenario in the database. Or any other object that has an attribute
        ``id`` that is an integer. If no :class:`eflips.model.Scenario` object is passed, the ``database_url``
        parameter must be set to a valid database URL ot the environment variable ``DATABASE_URL`` must be set to a
        valid database URL.

    :param repetition_period: An optional timedelta object specifying the period of the vehicle schedules. This
        is needed because the result should be a steady-state result. THis can only be achieved by simulating a
        time period before and after our actual simulation, and then only using the "middle". eFLIPS tries to
        automatically detect whether the schedule should be repeated daily or weekly. If this fails, a ValueError is
        raised and repetition needs to be specified manually.

    :param database_url: An optional database URL. If no database URL is passed and the `scenario` parameter is not a
        :class:`eflips.model.Scenario` object, the environment variable `DATABASE_URL` must be set to a valid database
        URL.

    :param smart_charging_strategy: An optional parameter specifying the smart charging strategy to be used. The
        default is SmartChargingStragegy.NONE. The following strategies are available:
        - SmartChargingStragegy.NONE: Do not use smart charging. Buses are charged with the maximum power available,
        from the time they arrive at the depot until they are full (or leave the depot).
        - SmartChargingStragegy.EVEN: Use smart charging with an even distribution of charging power over the time the
        bus is at the depot. This aims to minimize the peak power demand.

    :return: Nothing. The results are added to the database.
    """
    with create_session(scenario, database_url) as (session, scenario):
        simulation_host = init_simulation(
            scenario=scenario,
            session=session,
            repetition_period=repetition_period,
        )
        ev = run_simulation(simulation_host)
        add_evaluation_to_database(scenario, ev, session)

        match smart_charging_strategy:
            case SmartChargingStrategy.NONE:
                pass
            case SmartChargingStrategy.EVEN:
                apply_even_smart_charging(scenario, database_url)
            case SmartChargingStrategy.MIN_PRICE:
                raise NotImplementedError("MIN_PRICE strategy is not implemented yet.")
            case _:
                raise NotImplementedError()


def init_simulation(
    scenario: Scenario,
    session: Session,
    repetition_period: Optional[timedelta] = None,
    vehicle_count_dict: Optional[Dict[str, Dict[str, int]]] = None,
) -> SimulationHost:
    """
    This methods checks the input data for consistency, initializes a simulation host object and returns it.

    The simulation host object can then be passed to :func:`run_simulation()`.

    :param scenario: A :class:`eflips.model.Scenario` object containing the input data for the simulation.

    :param session: A SQLAlchemy session object.

    :param repetition_period: An optional timedelta object specifying the period of the vehicle schedules. This
        is needed because the *result* should be a steady-state result. THis can only be achieved by simulating a
        time period before and after our actual simulation, and then only using the "middle". eFLIPS tries to
        automatically detect whether the schedule should be repeated daily or weekly. If this fails, a ValueError is
        raised and repetition needs to be specified manually.

    :param vehicle_count_dict: An optional dictionary specifying the number of vehicles for each vehicle type for each
        depot. The dictionary should have the following structure:

         ::

            {
                "1" (depot.id as str): {
                    "1" (vehicle_type.id as str): 10,
                    "2" (vehicle_type.id as str): 20,
                    ...
                },
                "2" (depot.id as str): {
                    "1" (vehicle_type.id as str): 10,
                    "2" (vehicle_type.id as str): 20,
                    ...
                },

    :return: A :class:`eflips.depot.Simulation.SimulationHost` object. This object should be reagrded as a "black box"
        by the user. It should be passed to :func:`run_simulation()` to run the simulation and obtain the results.
    """

    # Clear the eflips settings
    eflips.settings.reset_settings()

    path_to_this_file = os.path.dirname(__file__)

    # Step 1: Set up the depot
    eflips_depots = []
    for depot in session.query(Depot).filter(Depot.scenario_id == scenario.id).all():
        # Step 1.5: Check validity of a depot
        check_depot_validity(depot)

        depot_dict = depot_to_template(depot)
        eflips_depots.append(
            eflips.depot.Depotinput(filename_template=depot_dict, show_gui=False)
        )

    # Step 2: eFLIPS initialization
    simulation_host = SimulationHost(eflips_depots, print_timestamps=False)
    # Now we do what is done in `standard_setup()` from the old input files
    # Load the settings
    path_to_default_settings = os.path.join(
        path_to_this_file, "defaults", "default_settings"
    )
    eflips.load_settings(path_to_default_settings)

    # Step 3: Set up the vehicle schedules
    # Turn rotations into vehicleschedules
    # Turn API VehicleSchedule objects into eFLIPS TimeTable object
    # Get correctly repeated vehicle schedules
    # if total duration time is 1 or 2 days, vehicle schedule will be repeated daily
    # if total duration time is 7 or 8 days, vehicle schedule will be repeated weekly
    # However, we need to override quite a lot of the settings
    # The ["general"]["SIMULATION_TIME"] entry is calculated from the difference between the first and last departure
    # time in the vehicle schedule
    vehicle_schedules = [
        VehicleSchedule.from_rotation(
            rotation,
            scenario=scenario,
            session=session,
        )
        for rotation in session.query(Rotation)
        .filter(Rotation.scenario_id == scenario.id)
        .all()
    ]

    first_departure_time = min(
        [vehicle_schedule.departure for vehicle_schedule in vehicle_schedules]
    )
    last_arrival_time = max(
        [vehicle_schedule.arrival for vehicle_schedule in vehicle_schedules]
    )

    # We take first arrival time as simulation start
    total_duration = (last_arrival_time - first_departure_time).total_seconds()
    schedule_duration_days = ceil(total_duration / (24 * 60 * 60))

    if repetition_period is None and schedule_duration_days in [1, 2]:
        repetition_period = timedelta(days=1)
    elif repetition_period is None and schedule_duration_days in [7, 8]:
        repetition_period = timedelta(weeks=1)
    elif repetition_period is None:
        raise ValueError(
            "Could not automatically detect repetition period. Please specify manually."
        )

    # Now, we need to repeat the vehicle schedules

    vehicle_schedules = repeat_vehicle_schedules(vehicle_schedules, repetition_period)

    sim_start_stime, total_duration_seconds = start_and_end_times(vehicle_schedules)
    eflips.globalConstants["general"]["SIMULATION_TIME"] = int(total_duration_seconds)

    eflips.globalConstants["general"]["SIMULATION_START_DATETIME"] = sim_start_stime

    timetable = VehicleSchedule._to_timetable(
        vehicle_schedules, simulation_host.env, sim_start_stime
    )
    simulation_host.timetable = timetable

    # Step 4: Set up the vehicle types
    # Clear old vehicle counts, if they exist
    eflips.globalConstants["depot"]["vehicle_count"] = {}

    # We need to calculate roughly how many vehicles we need for each depot
    for depot in session.query(Depot).filter(Depot.scenario_id == scenario.id).all():
        depot_id = str(depot.id)
        eflips.globalConstants["depot"]["vehicle_count"][depot_id] = {}
        vehicle_types_for_depot = set(str(area.vehicle_type_id) for area in depot.areas)
        if "None" in vehicle_types_for_depot:
            vehicle_types_for_depot.remove("None")

        # If we have a vehicle count dictionary, we validate and use ir
        if vehicle_count_dict is not None and depot_id in vehicle_count_dict.keys():
            if set(vehicle_count_dict[depot_id].keys()) < vehicle_types_for_depot:
                raise ValueError(
                    "The vehicle count dictionary does not contain all vehicle types for depot {depot_id}."
                )
            eflips.globalConstants["depot"]["vehicle_count"][
                depot_id
            ] = vehicle_count_dict[depot_id]
        else:
            # Calculate it from the size of the charging area with a 2x margin

            for vehicle_type in vehicle_types_for_depot:
                vehicle_count = 0
                for area in depot.areas:
                    if area.vehicle_type_id == int(vehicle_type):
                        # TODO potential edit if we make vehicle type of an area a list
                        for p in area.processes:
                            if p.electric_power is not None and p.duration is None:
                                vehicle_count += area.capacity

                assert (
                    vehicle_count > 0
                ), f"The charging area capacity for vehicle type {vehicle_type} should not be 0."

                eflips.globalConstants["depot"]["vehicle_count"][depot_id][
                    vehicle_type
                ] = (vehicle_count * 2)

    # We  need to put the vehicle type objects into the GlobalConstants
    for vehicle_type in (
        session.query(VehicleType).filter(VehicleType.scenario_id == scenario.id).all()
    ):
        eflips.globalConstants["depot"]["vehicle_types"][
            str(vehicle_type.id)
        ] = vehicle_type_to_global_constants_dict(vehicle_type)

    # Step 5: Final checks and setup
    # Run the eflips validity checks
    eflips.depot.settings_config.check_gc_validity()

    # Complete the eflips settings
    eflips.depot.settings_config.complete_gc()

    # Set up the depots
    for dh, di in zip(simulation_host.depot_hosts, simulation_host.to_simulate):
        dh.load_and_complete_template(di.filename_template)

    simulation_host.complete()

    return simulation_host


def run_simulation(simulation_host: SimulationHost) -> Dict[str, DepotEvaluation]:
    """Run simulation and return simulation results.

    :param simulation_host: A "black box" object containing all input data for the simulation.

    :return: A dictionary of :class:`eflips.depot.evaluation.DepotEvaluation` objects. The keys are the depot IDs, as
        strings.
    """
    simulation_host.run()

    results = {}
    for depot_host in simulation_host.depot_hosts:
        depot_id = depot_host.depot.ID
        ev = depot_host.evaluation

        # We need to clean up the timetable, it has trips from all depots
        ev.timetable = copy.copy(ev.timetable)
        ev.timetable.trips = [
            t for t in ev.timetable.trips if t.destination.ID == depot_id
        ]
        ev.timetable.trips_issued = [
            t for t in ev.timetable.trips_issued if t.destination.ID == depot_id
        ]
        ev.timetable.all_trips = [
            t for t in ev.timetable.all_trips if t.destination.ID == depot_id
        ]

        # We also need to clean up the vehicle generator
        ev.vehicle_generator = copy.copy(ev.vehicle_generator)
        ev.vehicle_generator.items = copy.copy(ev.vehicle_generator.items)
        indizes_to_remove = []
        for i in range(len(ev.vehicle_generator.items)):
            vehicle = copy.copy(ev.vehicle_generator.items[i])
            depot_ids = [trip.destination.ID for trip in vehicle.finished_trips]
            if len(set(depot_ids)) > 1:
                raise ValueError("Vehicle has finished trips in multiple depots.")
            if len(depot_ids) == 0 or depot_ids[0] != depot_id:
                indizes_to_remove.append(i)

        for index in sorted(indizes_to_remove, reverse=True):
            del ev.vehicle_generator.items[index]

        results[depot_id] = ev

    return results


def insert_dummy_standby_departure_events(
    depot_id: int, session: Session, sim_time_end: Optional[datetime.datetime] = None
) -> None:
    """
    Workaround for the missing STANDBY_DEPARTURE events in the database.

    :param session: The database session
    :param scenario: A scenario object
    :param sim_time_end: The end time of the simulation. If None, final events might not be properly handled.
    :return:
    """
    logger = logging.getLogger(__name__)

    # Look for charging events at areas belonging to the depot
    charging_events = (
        session.query(Event)
        .join(Area)
        .filter(Area.depot_id == depot_id)
        .filter(Event.event_type == EventType.CHARGING_DEPOT)
        .all()
    )

    for charging_event in charging_events:
        # See if the next event is a DRIVING event, but there is time between the two events
        next_event = (
            session.query(Event)
            .filter(Event.time_start >= charging_event.time_end)
            .filter(Event.vehicle_id == charging_event.vehicle_id)
            .order_by(Event.time_start)
            .first()
        )
        if (
            next_event is not None
            and next_event.event_type == EventType.DRIVING
            and (next_event.time_start - charging_event.time_end) > timedelta(seconds=1)
        ):
            logger.warning("Inserting dummy STANDBY_DEPARTURE event")
            # Insert a dummy STANDBY_DEPARTURE event
            dummy_event = Event(
                vehicle_id=charging_event.vehicle_id,
                vehicle_type_id=charging_event.vehicle.vehicle_type_id,
                time_start=charging_event.time_end,
                time_end=(next_event.time_start - timedelta(seconds=1)),
                event_type=EventType.STANDBY_DEPARTURE,
                area_id=charging_event.area_id,
                subloc_no=charging_event.subloc_no,
                scenario_id=charging_event.scenario_id,
                soc_start=charging_event.soc_end,
                soc_end=charging_event.soc_end,
                description="Dummy STANDBY_DEPARTURE event",
            )
            session.add(dummy_event)
        elif next_event is None and sim_time_end is not None:
            # If the event's end is before the simulation end, insert a dummy STANDBY_DEPARTURE event
            # From the end of the charging event to the end of the simulation
            logger.warning("Inserting dummy STANDBY_DEPARTURE event")
            if charging_event.time_end < sim_time_end:
                dummy_event = Event(
                    vehicle_id=charging_event.vehicle_id,
                    vehicle_type_id=charging_event.vehicle.vehicle_type_id,
                    time_start=charging_event.time_end,
                    time_end=sim_time_end,
                    event_type=EventType.STANDBY_DEPARTURE,
                    area_id=charging_event.area_id,
                    subloc_no=charging_event.subloc_no,
                    scenario_id=charging_event.scenario_id,
                    soc_start=charging_event.soc_end,
                    soc_end=charging_event.soc_end,
                    description="Dummy STANDBY_DEPARTURE event",
                )
                session.add(dummy_event)


def add_evaluation_to_database(
    scenario: Scenario,
    depot_evaluations: Dict[str, DepotEvaluation],
    session: sqlalchemy.orm.Session,
) -> None:
    """
    This method adds a simulation result to the database.

    It reads the simulation results from the :class:`eflips.depot.evaluation.DepotEvaluation` object and  adds them into
     the database. Tables of Event, Rotation and Vehicle will be updated.

    :param scenario: A :class:`eflips.model.Scenario` object containing the input data for the simulation.

    :param depot_evaluations: A dictionary of :class:`eflips.depot.evaluation.DepotEvaluation` objects. The keys are
        the depot IDs, as strings.

    :param session: a SQLAlchemy session object. This is used to add all the simulation results to the
        database.

    :return: Nothing. The results are added to the database.
    """

    # Read simulation start time

    for depot_id, depot_evaluation in depot_evaluations.items():
        simulation_start_time = depot_evaluation.sim_start_datetime

        # Depot-layer operations

        list_of_assigned_schedules = []

        waiting_area_id = None

        total_areas = session.query(Area).filter(Area.scenario_id == scenario.id).all()
        for area in total_areas:
            if area.depot_id == int(depot_id) and len(area.processes) == 0:
                waiting_area_id = area.id

        assert isinstance(waiting_area_id, int) and waiting_area_id > 0, (
            f"Waiting area id should be an integer greater than 0. For every depot there must be at least "
            f"one waiting area."
        )

        for current_vehicle in depot_evaluation.vehicle_generator.items:
            # Vehicle-layer operations

            vehicle_type_id = int(current_vehicle.vehicle_type.ID)

            current_vehicle_db = Vehicle(
                vehicle_type_id=vehicle_type_id,
                scenario=scenario,
                name=current_vehicle.ID,
                name_short=None,
            )

            # Flush the vehicle object to get the vehicle id
            session.add(current_vehicle_db)
            session.flush()

            dict_of_events = OrderedDict()

            (
                schedule_current_vehicle,
                earliest_time,
                latest_time,
                # Earliest and latest time defines a time window, only the events within this time window will be
                # handled. It is usually the departure time of the last copy trip in the "early-shifted" copy
                # schedules and the departure time of the first copy trip in the "late-shifted" copy schedules.
            ) = get_finished_schedules_per_vehicle(
                dict_of_events, current_vehicle.finished_trips, current_vehicle_db.id
            )

            try:
                assert earliest_time is not None and latest_time is not None

            except AssertionError as e:
                warnings.warn(
                    f"Vehicle {current_vehicle_db.id} has only copied trips. The profiles of this vehicle "
                    f"will not be written into database."
                )
                continue

            assert (
                earliest_time < latest_time
            ), f"Earliest time {earliest_time} is not less than latest time {latest_time}."

            list_of_assigned_schedules.extend(schedule_current_vehicle)

            generate_vehicle_events(
                dict_of_events,
                current_vehicle,
                waiting_area_id,
                earliest_time,
                latest_time,
            )

            # Python passes dictionaries by reference

            complete_standby_departure_events(dict_of_events, latest_time)

            add_soc_to_events(dict_of_events, current_vehicle.battery_logs)

            try:
                assert (not dict_of_events) is False
            except AssertionError as e:
                warnings.warn(
                    f"Vehicle {current_vehicle_db.id} has no valid events. The vehicle will not be written "
                    f"into database."
                )

                continue

            add_events_into_database(
                current_vehicle_db,
                dict_of_events,
                session,
                scenario,
                simulation_start_time,
            )

        # Postprocessing of events
        update_vehicle_in_rotation(session, scenario, list_of_assigned_schedules)
        update_waiting_events(session, scenario, waiting_area_id)



"""Utilities for defining a depot layout as a packing problem and solving it.
"""
import math
from decimal import Decimal
from random import randint
from operator import attrgetter
import itertools
from copy import copy
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.animation import FuncAnimation

language = "de"  # language for plots and animation, 'en' or 'de'


# Definition of constant depot layout parameters. Unit is metres unless
# otherwise stated.


# Parameters for SB (12m bus)
WIDTH = 2.55
LENGTH = 12

WIDTH_SAFE = 3.55  # n
LENGTH_SAFE = 12.5  # m
ANGLE_DIRECT = 45  # degrees (int between 0 and 75)

DIRECT_DISTANCE_A = 8
DIRECT_DISTANCE_B = 0

LINE_DISTANCE_A = 0
LINE_DISTANCE_B = 19.25

EDGE_DISTANCE_A = 8
EDGE_DISTANCE_B = 15


# Parameters for AB (18m bus)

# WIDTH = 2.55
# LENGTH = 18
#
# WIDTH_SAFE = 3.55   # n
# LENGTH_SAFE = 18.5  # m
# ANGLE_DIRECT = 45   # degrees (int between 0 and 75)
#
# DIRECT_DISTANCE_A = 10
# DIRECT_DISTANCE_B = 0
#
# LINE_DISTANCE_A = 0
# LINE_DISTANCE_B = 19.25
#
# EDGE_DISTANCE_A = 8
# EDGE_DISTANCE_B = 15


class ValidationError(Exception):
    def __init__(self, message):
        self.message = message


class Rectangle:
    """Rectangle representation as basis for subclasses.

    a and b: width and height
    x and y: coordinates of bottom left corner
    """

    def __init__(
        self,
        a,
        b,
        angle=0,
        x=0,
        y=0,
        fill=True,
        color="black",
        alpha=1.0,
        linestyle="-",
        **kwargs
    ):
        self.a = Decimal(str(a))
        self.b = Decimal(str(b))
        self.angle = int(angle)
        self._x = Decimal(str(x))
        self._y = Decimal(str(y))

        self.fill = fill
        self.color = color
        self.alpha = alpha
        self.linestyle = linestyle

    def __repr__(self):
        return "{%s} a=%s, b=%s, x=%s, y=%s, A=%s" % (
            type(self).__name__,
            self.a,
            self.b,
            self.x,
            self.y,
            self.A,
        )

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        self._x = Decimal(str(value))

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        self._y = Decimal(str(value))

    @property
    def A(self):
        return self.a * self.b

    @property
    def x_left(self):
        """Return x of the left side. Must be unrotated."""
        return self.x

    @property
    def x_right(self):
        """Return x of the right side. Must be unrotated."""
        return self.x + self.a

    @property
    def y_top(self):
        """Return y of the top side. Must be unrotated."""
        return self.y + self.b

    @property
    def y_bottom(self):
        """Return y of the bottom side. Must be unrotated."""
        return self.y

    @property
    def x_tr(self):
        """Return x of the top right corner. Must be unrotated."""
        return self.x + self.a

    @property
    def y_tr(self):
        """Return y of the top right corner. Must be unrotated."""
        return self.y + self.b

    @property
    def x_tl(self):
        """Return x of the top left corner. Must be unrotated."""
        return self.x

    @property
    def y_tl(self):
        """Return y of the top left corner. Must be unrotated."""
        return self.y + self.b

    @property
    def x_br(self):
        """Return x of the bottom right corner. Must be unrotated."""
        return self.x + self.a

    @property
    def y_br(self):
        """Return y of the bottom right corner. Must be unrotated."""
        return self.y

    @property
    def x_center(self):
        """Return x of the center. Must be unrotated."""
        return self.x + self.a / 2

    @property
    def y_center(self):
        """Return y of the center. Must be unrotated."""
        return self.y + self.b / 2


def fits_into(r1, r2):
    """Return True if rectangle r1 is smaller or equal than rectangle r2 in
    both dimensions.
    """
    return r1.a <= r2.a and r1.b <= r2.b


def xintersect(r1, r2):
    """Return True if unrotated rectangles r1 and r2 intersect in x direction.
    Same value is not considered intersection.
    """
    return r1.x_left < r2.x_right and r1.x_right > r2.x_left


def yintersect(r1, r2):
    """Return True if unrotated rectangles r1 and r2 intersect in y direction.
    Same value is not considered intersection.
    """
    return r1.y_bottom < r2.y_top and r1.y_top > r2.y_bottom


def intersect(r1, r2):
    """Return True if unrotated rectangles r1 and r2 intersect. Same value is
    not considered intersection.
    """
    return xintersect(r1, r2) and yintersect(r1, r2)


def contains(r1, r2):
    """Return True if rectangle r1 completely encloses rectangle r2. Must be
    unrotated.
    """
    return all(
        [
            r1.x_left <= r2.x_left,
            r1.y_bottom <= r2.y_bottom,
            r1.x_right >= r2.x_right,
            r1.y_top >= r2.y_top,
        ]
    )


def contains_point(r, p):
    """Return True if rectangle r contains point p. Must be unrotated."""
    return all([r.x_left <= p.x <= r.x_right, r.y_bottom <= p.y <= r.y_top])


class RectangleWithInner(Rectangle):
    """Rectangle containing two or more identical rectangles.
    Inner rectangles are stacked in the direction of increasing y.

    m, n: [int or float] a, b of the inner rectangles
    """

    def __init__(self, m, n, count_inner, x=0, y=0, **kwargs):
        if not str(count_inner).isdigit() or count_inner < 2:
            raise ValueError("count_inner must be a natural number > 1.")
        self.count_inner = count_inner
        self.inner = Rectangle(m, n, 0, x, y, color="cornflowerblue")

        super(RectangleWithInner, self).__init__(a=m, b=count_inner * n, x=x, y=y)

        self.fill = False

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        value = Decimal(str(value))
        self.inner.x = value
        self._x = value

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        value = Decimal(str(value))
        self.inner.y = value
        self._y = value

    @property
    def A_inner(self):
        """Return the total space of inner rectangles."""
        return self.inner.A * self.count_inner

    @property
    def util_rate(self):
        return self.A_inner / self.A

    def x_inner(self, *args):
        """Return the x coordinate of inner rectangles (same for all)."""
        return self.inner.x

    def y_inner(self, index):
        """Return the y coordinate of the inner rectangle with *index*
        (starting at 0).
        """
        return self.inner.y + self.inner.b * index


class RectangleWithRotatableInner(Rectangle):
    """Rectangle containing one or more identical rotatable rectangles.
    Inner rectangles are stacked in the direction of increasing y.

    Parameters:
    m, n: [int or float] a, b of the inner rectangles

    Attributes:
    inner: [Rectangle] bottom inner rectangle

    """

    def __init__(self, m, n, count_inner=1, angle_inner=45, x=0, y=0, **kwargs):
        self.h = Decimal(str(0))
        self.inner = Rectangle(m, n, angle_inner, x, y, color="orange")
        if not str(count_inner).isdigit() or count_inner < 1:
            raise ValueError("'count_inner' must be a natural number > 0.")
        self.count_inner = count_inner

        super(RectangleWithRotatableInner, self).__init__(a=0, b=0, x=x, y=y)

        self.rotate_inner(angle_inner)
        self.fill = False

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        value = Decimal(str(value))
        # Move inner x by difference because its not equal to self.x
        self.inner.x += value - self._x
        self._x = value

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        value = Decimal(str(value))
        self.inner.y += value - self._y
        self._y = value

    @property
    def A_inner(self):
        """Return the total space of inner rectangles."""
        return self.inner.A * self.count_inner

    @property
    def util_rate(self):
        return self.A_inner / self.A

    def rotate_inner(self, angle):
        """Set the rotation of the inner rectangles and update the parameters
        of self.

        A reference drawing for names used in this method such as m, AQ is
        located at depot_layout_opt/doc/direct_details.pdf.

        angle: [int or float] value between -75 and 75
        """
        if not str(abs(angle)).isdigit() or not 0 <= abs(angle) <= 75:
            raise ValueError("angle_inner must be between -75 " "and 75.")
        angle_rad = math.radians(angle)
        m = self.inner.a
        n = self.inner.b
        PA = n * Decimal(str(math.sin(angle_rad)))  # line PA in drawing
        AQ = m * Decimal(str(math.cos(angle_rad)))  # line AQ in drawing

        self.h = n / Decimal(str(math.cos(angle_rad)))
        PD = n * Decimal(str(math.cos(angle_rad)))  # line PD in drawing
        DS = m * Decimal(str(math.sin(angle_rad)))  # line DS in drawing

        self.inner.angle = angle

        if angle < 0:
            self.a = -PA + AQ
            self.b = PD - DS + (self.count_inner - 1) * self.h
            self.inner.x = self.x
            self.inner.y = self.y - DS
        else:
            self.a = PA + AQ
            self.b = PD + DS + (self.count_inner - 1) * self.h
            self.inner.x = self.x + PA

    def x_inner(self, *args):
        """Return the x coordinate of inner rectangles (same for all)."""
        return self.inner.x

    def y_inner(self, index):
        """Return the y coordinate of the inner rectangle with *index*
        (starting at 0).
        """
        return self.inner.y + index * self.h


class RectangleWithRotatedDoubleRowInner(Rectangle):
    """Rectangle containing two or more identical sized rectangles that
    are rotated by 45° and arranged in a double row.

    Parameters:
    m, n: [int or float] a, b of the inner rectangles

    Attributes:
    inner: [Rectangle] bottom inner rectangle

    """

    def __init__(self, m, n, count_inner=2, x=0, y=0, **kwargs):
        self.h = Decimal(str(0))
        self.inner_left = Rectangle(m, n, 45, x, y, color="orange")
        self.inner_right = Rectangle(m, n, 135, x, y, color="orange")
        if not str(count_inner).isdigit() or count_inner < 2:
            raise ValueError("'count_inner' must be a natural number > 1.")
        self.count_inner = count_inner

        super(RectangleWithRotatedDoubleRowInner, self).__init__(a=0, b=0, x=x, y=y)
        self.fill = False

        # Update values from rotating the inner rectangles to 45° for left
        # and 135° for right.
        # A reference drawing for names used below such as m, PA ism located at
        # depot_layout_opt/doc/direct_details.pdf.

        uneven_count = count_inner % 2
        angle_rad = math.radians(45)
        m = self.inner_left.a  # reassign due to conversion to decimal
        n = self.inner_left.b

        PA = n * Decimal(str(math.sin(angle_rad)))  # line PD in drawing
        ABx = m * Decimal(str(math.cos(angle_rad)))  # x value of line AQ in drawing
        self.a = PA + 2 * ABx
        self.h = n / Decimal(str(math.cos(angle_rad)))
        PD = n * Decimal(str(math.cos(angle_rad)))  # line PD in drawing
        DS = m * Decimal(str(math.sin(angle_rad)))  # line DS in drawing
        n_slots_left = Decimal(str(self.count_inner / 2)) + int(
            not uneven_count
        ) * Decimal(str(0.5))
        self.b = PD + DS + (n_slots_left - 1) * self.h + uneven_count * self.h / 2

        self.inner_left.x = self.x + PA
        self.inner_right.x = self.x + self.a
        self.inner_right.y = y + self.h

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        value = Decimal(str(value))
        # Move inner x by difference because its not equal to self.x
        self.inner_left.x += value - self._x
        self.inner_right.x += value - self._x
        self._x = value

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        value = Decimal(str(value))
        self.inner_left.y = value
        # Move inner_right y by difference because its not equal to self.x
        self.inner_right.y += value - self._y
        self._y = value

    @property
    def A_inner(self):
        """Return the total space of inner rectangles."""
        return self.inner_left.A * self.count_inner

    @property
    def util_rate(self):
        return self.A_inner / self.A

    def x_inner(self, index):
        """Return the x coordinate of inner rectangle with *index*. Indexing
        starts at zero, alternating left and right from the bottom left towards
        the top.
        """
        if index % 2:
            return self.inner_right.x
        else:
            return self.inner_left.x

    def y_inner(self, index):
        """Return the y coordinate of inner rectangle with *index*. Indexing
        starts at zero, alternating left and right from the bottom left towards
        the top.
        """
        if index % 2:
            return self.inner_right.y + Decimal(str((index - 1) / 2)) * self.h
        else:
            return self.inner_left.y + Decimal(str(index / 2)) * self.h


class Available(Rectangle):
    """Rectangle with default parameters to represent available space in
    a bin.
    """

    def __init__(
        self, a, b, x=0, y=0, fill=False, color="black", alpha=0.7, linestyle="--"
    ):
        super(Available, self).__init__(a, b, 0, x, y, fill, color, alpha, linestyle)


def alab_arib(av, item):
    """Return a new [Available] spanning from (av.x_left, av.y_bottom) to
    (av.x_right, item.y_bottom).
    """
    return Available(
        a=av.x_right - av.x_left,
        b=item.y_bottom - av.y_bottom,
        x=av.x_left,
        y=av.y_bottom,
    )


def alit_arat(av, item):
    """Return a new [Available] spanning from (av.x_left, item.y_top) to
    (av.x_right, av.y_top).
    """
    return Available(
        a=av.x_right - av.x_left, b=av.y_top - item.y_top, x=av.x_left, y=item.y_top
    )


def alab_ilat(av, item):
    """Return a new [Available] spanning from (av.x_left, av.y_bottom) to
    (item.x_left, av.y_top).
    """
    return Available(
        a=item.x_left - av.x_left, b=av.y_top - av.y_bottom, x=av.x_left, y=av.y_bottom
    )


def irab_arat(av, item):
    """Return a new [Available] spanning from (item.x_right, av.y_bottom) to
    (av.x_right, av.y_top).
    """
    return Available(
        a=av.x_right - item.x_right,
        b=av.y_top - av.y_bottom,
        x=item.x_right,
        y=av.y_bottom,
    )


class Bin(Rectangle):
    """Rectangular container with best-fit-decreasing algorithm for rectangle
    packing.
    """

    def __init__(self, a, b, record_history=True, fill=False):
        super(Bin, self).__init__(a, b, fill=fill)
        self.items = []
        self.availables = []
        self.packed_items = []
        self.record_history = record_history
        self.history = {"items": [], "availables": []}
        self._feasible = None
        self._precheck_passed = None

        self.fig = None
        self.ax = None

        self.availables.append(Available(a, b, x=0, y=0))

    @property
    def valid(self):
        """Return True if none of the rectangles in self.packed_items
        intersect, else return False.
        """
        for combo in itertools.combinations(self.packed_items, 2):
            if intersect(*combo):
                return False
        return True

    @property
    def feasible(self):
        """Return the result of the packing attempt as boolean. None until
        self.pack is called.
        """
        return self._feasible

    @property
    def count_inner(self):
        """Return the total number of inner rectangles of items."""
        return sum(
            pitem.count_inner for pitem in self.items if hasattr(pitem, "count_inner")
        )

    @property
    def A_inner(self):
        """Return the total space of items in unpacked state. Excludes space
        for distances (if applicable).
        """
        return sum(item.A for item in self.items)

    @property
    def util_rate(self):
        """Percentage of the total bin space occupied by packed items."""
        used = sum(pitem.A for pitem in self.packed_items)
        return used / self.A

    @property
    def precheck_passed(self):
        """Return the result of the precheck as boolean. None until
        self.precheck is called.
        """
        return self._precheck_passed

    def precheck(self):
        """Quick checks applicable before packing to detect if items fit."""
        for item in self.items:
            if item.a > self.a:
                # print('Side a of an item is too large.')
                self._precheck_passed = False
                return
            if item.b > self.b:
                # print('Side b of an item is too large.')
                self._precheck_passed = False
                return
        if sum(item.A for item in self.items) > self.A:
            # print('The sum of areas of items is larger than the available '
            #       'area.')
            self._precheck_passed = False
            return

        import itertools

        combos = itertools.combinations(self.items, 2)
        for i1, i2 in combos:
            assert i1 is not i2

        self._precheck_passed = True

    def pack(self):
        """Try to distribute all items into self so that none intersect and
        return True if successful, else return False.
        """
        if self.packed_items:
            raise RuntimeError("Cannot pack again without reset.")
        if not self.items:
            raise RuntimeError("Cannot pack because item list is emtpy.")

        self.precheck()
        if not self.precheck_passed:
            self._feasible = False
            return

        # Simplified sorting: Sort by 'a' (primary), then by 'b'.
        # 'conflict_category' is not necessary in the depot scenario as long as
        # 'a' yields the same result.
        self.items.sort(key=attrgetter("conflict_category", "a", "b"), reverse=True)
        # items.sort(key=attrgetter('conflict_category', 'a', 'b'), reverse=True)

        if self.record_history:
            self.history["items"].append(self.packed_items.copy())
            self.history["availables"].append(self.availables.copy())

        for item in self.items:
            # print('Next item: %s' % item)
            if not self.put(item):
                self._feasible = False
                return

            if self.record_history:
                self.history["items"].append(self.packed_items.copy())
                self.history["availables"].append(self.availables.copy())

        assert self.valid
        self._feasible = True
        # print('packed')

    def put(self, item):
        """Call try_put to determine if item fits into self. If successful,
        execute the action and return True, else False.
        """
        success, av_put = self.try_put(item)

        if success:
            self.packed_items.append(item)
            self.update_availables(item)

        return success

    def try_put(self, item):
        """Determine if there is valid available space for item. Return (True,
        the available space rectangle) if yes, else (False, None). x and y of
        item may be manipulated regardless of the
        result.
        """
        for av in self.availables:
            if av.a >= item.a and av.b >= item.b:
                # success
                item.x = av.x
                item.y = av.y
                return True, av
        return False, None

    def update_availables(self, item):
        """Update availables after packing *item*."""
        # Split existing availables that intersect item into several smaller
        # ones depending on the relative location
        new_avs = []
        removals = []
        for av in self.availables:
            if intersect(av, item):
                removals.append(av)
                # Get differences of bottom left and top right corner
                # coordinates
                dx_bl = av.x_left - item.x_left
                dy_bl = av.y_bottom - item.y_bottom
                dx_tr = av.x_right - item.x_right
                dy_tr = av.y_top - item.y_top

                # Simplify differences to booleans
                bx_bl = dx_bl >= 0
                by_bl = dy_bl >= 0
                bx_tr = dx_tr > 0
                by_tr = dy_tr > 0

                if bx_bl:
                    if by_bl:
                        if bx_tr:
                            if by_tr:
                                new_avs.extend(self.case_8(av, item))
                            else:
                                new_avs.extend(self.case_12(av, item))
                        else:
                            if by_tr:
                                new_avs.extend(self.case_11(av, item))
                            else:
                                new_avs.extend(self.case_15(av, item))
                    else:
                        if bx_tr:
                            if by_tr:
                                new_avs.extend(self.case_1(av, item))
                            else:
                                new_avs.extend(self.case_2(av, item))
                        else:
                            if by_tr:
                                new_avs.extend(self.case_14(av, item))
                            else:
                                new_avs.extend(self.case_9(av, item))
                else:
                    if by_bl:
                        if bx_tr:
                            if by_tr:
                                new_avs.extend(self.case_7(av, item))
                            else:
                                new_avs.extend(self.case_13(av, item))
                        else:
                            if by_tr:
                                new_avs.extend(self.case_6(av, item))
                            else:
                                new_avs.extend(self.case_10(av, item))
                    else:
                        if bx_tr:
                            if by_tr:
                                new_avs.extend(self.case_16(av, item))
                            else:
                                new_avs.extend(self.case_3(av, item))
                        else:
                            if by_tr:
                                new_avs.extend(self.case_5(av, item))
                            else:
                                new_avs.extend(self.case_4(av, item))

        for av in removals:
            self.availables.remove(av)
        self.availables.extend(new_avs)

        # Remove availables that are fully enclosed by others and therefore
        # redundant
        removals = set()
        # Sort to be able to use combinations instead of permutations (less
        # pairs, therefore faster)
        pool = sorted(self.availables, key=lambda av: av.A, reverse=True)
        for r1, r2 in itertools.combinations(pool, 2):
            if contains(r1, r2):
                removals.add(r2)
                # print('Removing %s contained by %s' % (r1, r2))
        for av in removals:
            self.availables.remove(av)

        # Validation
        for av in self.availables:
            assert av.a > 0 and av.b > 0
            for pitem in self.packed_items:
                assert not intersect(av, pitem)
                # if intersect(av, pitem):
                #     raise ValidationError('Intersection of av: %s and item %s' % (av, pitem))

        self.availables.sort(key=attrgetter("a"))
        self.availables.sort(key=attrgetter("b"))

    @staticmethod
    def case_1(av, item):
        """Case 1 of splitting av into smaller rectangles."""
        return [alab_arib(av, item), alit_arat(av, item), irab_arat(av, item)]

    @staticmethod
    def case_2(av, item):
        return [alab_arib(av, item), irab_arat(av, item)]

    @staticmethod
    def case_3(av, item):
        return [alab_arib(av, item), alab_ilat(av, item), irab_arat(av, item)]

    @staticmethod
    def case_4(av, item):
        return [alab_arib(av, item), alab_ilat(av, item)]

    @staticmethod
    def case_5(av, item):
        return [alab_arib(av, item), alit_arat(av, item), alab_ilat(av, item)]

    @staticmethod
    def case_6(av, item):
        return [alit_arat(av, item), alab_ilat(av, item)]

    @staticmethod
    def case_7(av, item):
        return [alit_arat(av, item), alab_ilat(av, item), irab_arat(av, item)]

    @staticmethod
    def case_8(av, item):
        return [alit_arat(av, item), irab_arat(av, item)]

    @staticmethod
    def case_9(av, item):
        return [alab_arib(av, item)]

    @staticmethod
    def case_10(av, item):
        return [alab_ilat(av, item)]

    @staticmethod
    def case_11(av, item):
        return [alit_arat(av, item)]

    @staticmethod
    def case_12(av, item):
        return [irab_arat(av, item)]

    @staticmethod
    def case_13(av, item):
        return [alab_ilat(av, item), irab_arat(av, item)]

    @staticmethod
    def case_14(av, item):
        return [alab_arib(av, item), alit_arat(av, item)]

    @staticmethod
    def case_15(av, item):
        raise ValidationError("case 15 (ex)")

    @staticmethod
    def case_16(av, item):
        raise ValidationError("case 16 (bo)")

    def repack(self):
        """Reset packing values and pack again."""
        self.availables = []
        self.packed_items = []
        self.history = {"items": [], "availables": []}
        self._feasible = None
        self._precheck_passed = None

        self.fig = None
        self.ax = None

        self.availables.append(Rectangle(self.a, self.b, x=0, y=0))

        self.pack()

    def draw(self):
        """Plot self.packed_items."""
        if not self.packed_items:
            print("No packed items to draw.")
        return draw_rectangles_newplot(
            self.packed_items + [self], (0, float(self.a)), (0, float(self.b)), False
        )

    def save_drawing(
        self, filename, formats=("pdf",), confirm=True, dpi=None, show=False, **kwargs
    ):
        """Save what self.draw shows.

        filename: [str] including path, excluding extension. Existing files
            with the same name are overwritten without confirmation prompt.
        formats: [tuple] of file extensions [str]
        dpi: Parameter of plt.savefig()
        show: [bool] if False, the fig is closed automatically
        """
        fig, ax = self.draw()
        if "png" in formats:
            fig.savefig(filename + ".png", dpi=dpi, **kwargs)
            if confirm:
                print("Saved %s.png" % filename)
        if "pdf" in formats:
            fig.savefig(filename + ".pdf", dpi=dpi, **kwargs)
            if confirm:
                print("Saved %s.pdf" % filename)

        if not show:
            plt.close(fig)

    def animate(self):
        anim = self._animate(draw_distances=False)
        return anim

    def _animate(self, draw_distances):
        """Based on https://stackoverflow.com/a/49382421"""
        if not self.record_history or not self.history:
            print(
                "animate requires record_history to be be True and at least "
                "one entry in history."
            )
            return None

        def get_sub():
            """Generator to get the next in availables of the current frame."""
            imax = len(anim.frame_seq_subs[anim.i]) - 1
            anim.i_sub = 0
            yield anim.frame_seq_subs[anim.i][anim.i_sub]
            while True:
                anim.i_sub += anim.direction_sub
                if anim.i_sub < 0:
                    anim.i_sub = imax
                elif anim.i_sub > imax:
                    anim.i_sub = 0
                yield anim.frame_seq_subs[anim.i][anim.i_sub]

        def get_specific(i):
            recs = anim.frame_seq_items[i].copy()
            subs = anim.frame_seq_subs[i].copy()
            recs.extend(subs)
            return recs

        def get_next():
            """Generator to determine the frame and get its next main
            rectangles and all availables.
            """
            imax = len(anim.frame_seq_items) - 1
            anim.i = 0
            plt.title(tl("Step") + " " + str(anim.i))
            anim.subgen = get_sub()

            yield get_specific(anim.i)

            while True:
                anim.i += anim.direction
                if anim.i < 0:
                    anim.i = imax
                    plt.gca().texts.clear()
                elif anim.i > imax:
                    anim.i = 0
                    plt.gca().texts.clear()
                anim.subgen = get_sub()
                plt.title(tl("Step") + " " + str(anim.i))

                yield get_specific(anim.i)

        def on_press(event):
            """Action to execute on buttom press."""
            if event.key.isspace():
                if anim.running:
                    anim.event_source.stop()
                else:
                    anim.event_source.start()
                anim.running ^= True  # toggle through xor
            elif event.key == "left":
                anim.direction = -1
            elif event.key == "right":
                anim.direction = +1
            elif event.key == "up":
                anim.direction_sub = +1
            elif event.key == "down":
                anim.direction_sub = -1

            # Manually update the plot
            if event.key in ["left", "right"]:
                plt.gca().texts.clear()
                rectangles = next(anim.frame_seq)
                draw_rectangles(rectangles, draw_distances)

                # plt.title('')
                # from eflips.depot.evaluation import savefig
                # def save(fig, filename):
                #     savefig(fig, filename, formats=('png',), dpi=300)

                # save(plt.gcf(), tl('Step') + ' ' + str(anim.i) + ', _')

            if not anim.running and event.key in ["up", "down"]:
                plt.gca().texts.clear()
                rectangles = anim.frame_seq_items[anim.i].copy()
                av = copy(next(anim.subgen))
                av.fill = True
                av.color = "green"
                av.alpha = 0.5
                rectangles.append(av)
                rectangles.extend(anim.frame_seq_subs[anim.i])
                title = (
                    tl("Step")
                    + " "
                    + str(anim.i)
                    + ", "
                    + tl("av")
                    + " "
                    + tl("no")
                    + " "
                    + str(anim.i_sub + 1)
                )
                plt.title(title)
                draw_rectangles(rectangles, draw_distances)

                # plt.title('')
                # from eflips.depot.evaluation import savefig
                # def save(fig, filename):
                #     savefig(fig, filename, formats=('png',), dpi=300)

                # save(plt.gcf(), title)

        fig, ax = plt.subplots()
        fig.canvas.mpl_connect("key_press_event", on_press)
        anim = FuncAnimation(
            fig,
            draw_rectangles,
            frames=get_next,
            fargs=(draw_distances,),
            interval=1000,
            repeat=True,
            save_count=len(self.history["items"]),
        )
        anim.running = True
        anim.direction = +1
        anim.i = 0
        anim.i_sub = 0

        anim.frame_seq_items = self.history["items"].copy()
        # Add self to item history to also draw self (and the bin edge
        # distances for BinWithDistances) without manipulating the history.
        for entry in anim.frame_seq_items:
            entry.append(self)

        anim.frame_seq_subs = self.history["availables"]
        anim.subgen = None
        anim.direction_sub = +1

        ax.set_aspect("equal")
        plt.xlim(0, float(self.a))
        plt.ylim(0, float(self.b))

        dpi = fig.get_dpi()
        fig.set_size_inches(1920.0 / float(dpi), 948.0 / float(dpi))

        plt.show()

        return anim

    def save_animation(self, filename, confirm=True, fps=1.5, bitrate=1800, **kwargs):
        """Save what self.animate shows as .mp4.

        filename: [str] including path, excluding extension. Existing files
            with the same name are overwritten without confirmation prompt.

        Accepts arguments for ffmpeg writer init and FuncAnimation.save as
        kwargs.
        """
        # Set up formatting for the movie files
        Writer = matplotlib.animation.writers["ffmpeg"]
        writer = Writer(fps=fps, metadata=dict(artist="Me"), bitrate=bitrate, **kwargs)

        anim = self.animate()
        anim.save(filename + ".mp4", writer=writer, **kwargs)
        if confirm:
            print("Saved %s.mp4" % filename)


class DistanceRectangle(Rectangle):
    """Rectangle with default parameters to represent a buffer distance in
    packing with distances.
    """

    def __init__(
        self, a, b, x=0, y=0, fill=True, color="grey", alpha=0.5, linestyle=":"
    ):
        super(DistanceRectangle, self).__init__(
            a, b, 0, x, y, fill, color, alpha, linestyle
        )


class BinWithDistances(Bin):
    """Rectangular container with best-fit-decreasing-algorithm for rectangle
    packing with optional buffer distances between items as well as between
    items and the container edges.

    Distance handling based on https://doi.org/10.1051/ro/2012007
    """

    def __init__(self, a, b, record_history=True):
        super(BinWithDistances, self).__init__(a, b, record_history)

        self.distance_left_inner = DistanceRectangle(EDGE_DISTANCE_A, b, x=0, y=0)
        self.distance_bottom_inner = DistanceRectangle(a, EDGE_DISTANCE_B, x=0, y=0)
        self.distance_right_inner = DistanceRectangle(
            EDGE_DISTANCE_A, b, x=self.x_right - Decimal(str(EDGE_DISTANCE_A)), y=0
        )
        self.distance_top_inner = DistanceRectangle(
            a, EDGE_DISTANCE_B, x=0, y=self.y_top - Decimal(str(EDGE_DISTANCE_B))
        )

    def try_put(self, item):
        """Determine if there is valid available space with respect to buffer
        distances for item. Return (True, the Available rectangle) if
        yes, else (False, None). x and y of item may be manipulated regardless
        of the result.
        """
        for av in self.availables:
            if av.a >= item.a and av.b >= item.b:
                # print('Considering av %s' % av)
                # item fits without buffer distance. Assign preliminary position.
                item.x = av.x
                item.y = av.y

                # Left bin edge
                if item.distance_left.x_left < self.x_left or intersect(
                    item, self.distance_left_inner
                ):
                    oldx = item.x
                    distance = max(item.distance_left.a, self.distance_left_inner.a)
                    item.x = self.x_left + distance
                    assert oldx < item.x
                    if not contains(av, item):
                        continue

                # Bottom bin edge
                if item.distance_bottom.y_bottom < self.y_bottom or intersect(
                    item, self.distance_bottom_inner
                ):
                    oldy = item.y
                    distance = max(item.distance_bottom.b, self.distance_bottom_inner.b)
                    item.y = self.y_bottom + distance
                    assert oldy <= item.y
                    if not contains(av, item):
                        continue

                # Move item to the right and top until all buffer distance
                # conflicts on the left and bottom of *item* are resolved. The
                # validity is ignored here and checked later.
                left_ok = False
                bottom_ok = False
                i = -1
                invalid = False
                while not (left_ok and bottom_ok):
                    i += 1
                    if i > 50:
                        raise ValidationError("Feels like too many iterations")
                    # print(' ' * i + 'Checking left and right. Iteration=%s, item=%s. x=%s, y=%s' % (i, item.ID, item.x, item.y))

                    # Left
                    for pitem in self.packed_items:
                        if intersect(item.distance_left, pitem) or intersect(
                            item, pitem.distance_right
                        ):
                            oldx = item.x
                            distance = max(item.distance_left.a, pitem.distance_right.a)
                            item.x = pitem.x_right + distance

                            bottom_ok = False
                            assert oldx <= item.x
                            # print('--x Conflict with item %s. Moved %s to the right by %s. oldx: %s, newx: %s' % (pitem.ID, item.ID, distance, oldx, item.x))
                    if not contains(av, item):
                        invalid = True
                        break
                    left_ok = True

                    # Bottom
                    if not bottom_ok:
                        for pitem in self.packed_items:
                            if intersect(item.distance_bottom, pitem) or intersect(
                                item, pitem.distance_top
                            ):
                                oldy = item.y
                                distance = max(
                                    item.distance_bottom.b, pitem.distance_top.b
                                )
                                item.y = pitem.y_top + distance

                                left_ok = False
                                assert oldy <= item.y
                                # print(item)
                                # print(pitem)
                                # print(item.distance_bottom)
                                # print(pitem.distance_top)
                                # print('--y Conflict with item %s. Moved %s to the top by %s. pitem.y_top=%s, oldy: %f, newy: %s' % (pitem.ID, item.ID, distance, pitem.y_top, oldy, item.y))
                        if not contains(av, item):
                            invalid = True
                            break
                        bottom_ok = True

                if invalid:
                    continue

                # Checks that make this av invalid for item if True
                invalid = False

                # Intersection with other items
                for pitem in self.packed_items:
                    if intersect(item, pitem):
                        invalid = True
                        break
                if invalid:
                    continue

                # Right bin edge
                if item.distance_right.x_right > self.x_right or intersect(
                    item, self.distance_right_inner
                ):
                    # print('Bin conflict on the right side.')
                    continue

                # Right
                for pitem in self.packed_items:
                    if intersect(item.distance_right, pitem) or intersect(
                        item, pitem.distance_left
                    ):
                        # print('Item conflict on the right side.')
                        invalid = True
                        break
                if invalid:
                    continue

                # Top bin edge
                if item.distance_top.y_top > self.y_top or intersect(
                    item, self.distance_top_inner
                ):
                    # print('Bin conflict on the top side.')
                    continue

                # Top
                for pitem in self.packed_items:
                    if intersect(item.distance_top, pitem) or intersect(
                        item, pitem.distance_bottom
                    ):
                        # print('Item conflict on the top side.')
                        invalid = True
                        break
                if invalid:
                    continue

                # print('Suitable av found')
                return True, av
        # print('Couldnt find suitable av')
        return False, None

    @staticmethod
    def case_16(av, item):
        return [
            alab_arib(av, item),
            alit_arat(av, item),
            alab_ilat(av, item),
            irab_arat(av, item),
        ]

    def draw(self):
        """Plot self.packed_items and the distances."""
        if not self.packed_items:
            print("No packed items to draw.")
        return draw_rectangles_newplot(
            self.packed_items + [self], (0, float(self.a)), (0, float(self.b)), True
        )

    def animate(self):
        anim = self._animate(draw_distances=True)
        return anim


class DistanceLeft(DistanceRectangle):
    """Buffer rectangle on the outer left side of the *item* rectangle."""

    def __init__(self, a, item):
        super(DistanceLeft, self).__init__(a, 0)

        self.item = item

    @property
    def b(self):
        return self.item.b

    @b.setter
    def b(self, value):
        pass

    @property
    def x(self):
        return self.item.x - self.a

    @x.setter
    def x(self, value):
        raise RuntimeError("x of distance rectangle cannot be set directly.")

    @property
    def y(self):
        return self.item.y

    @y.setter
    def y(self, value):
        raise RuntimeError("y of distance rectangle cannot be set directly.")


class DistanceBottom(DistanceRectangle):
    """Buffer rectangle on the outer bottom side of the *item* rectangle."""

    def __init__(self, b, item):
        super(DistanceBottom, self).__init__(0, b)

        self.item = item

    @property
    def a(self):
        return self.item.a

    @a.setter
    def a(self, value):
        pass

    @property
    def x(self):
        return self.item.x

    @x.setter
    def x(self, value):
        raise RuntimeError("x of distance rectangle cannot be set directly.")

    @property
    def y(self):
        return self.item.y - self.b

    @y.setter
    def y(self, value):
        raise RuntimeError("y of distance rectangle cannot be set directly.")


class DistanceRight(DistanceRectangle):
    """Buffer rectangle on the outer right side of the *item* rectangle."""

    def __init__(self, a, item):
        super(DistanceRight, self).__init__(a, 0)

        self.item = item

    @property
    def b(self):
        return self.item.b

    @b.setter
    def b(self, value):
        pass

    @property
    def x(self):
        return self.item.x_br

    @x.setter
    def x(self, value):
        raise RuntimeError("x of distance rectangle cannot be set directly.")

    @property
    def y(self):
        return self.item.y_br

    @y.setter
    def y(self, value):
        raise RuntimeError("y of distance rectangle cannot be set directly.")


class DistanceTop(DistanceRectangle):
    """Buffer rectangle on the outer top side of the *item* rectangle."""

    def __init__(self, b, item):
        super(DistanceTop, self).__init__(0, b)

        self.item = item

    @property
    def a(self):
        return self.item.a

    @a.setter
    def a(self, value):
        pass

    @property
    def x(self):
        return self.item.x_tl

    @x.setter
    def x(self, value):
        raise RuntimeError("x of distance rectangle cannot be set directly.")

    @property
    def y(self):
        return self.item.y_tl

    @y.setter
    def y(self, value):
        raise RuntimeError("y of distance rectangle cannot be set directly.")


class RectangleWithDistances(Rectangle):
    """A rectangle that may have a buffer distance at each side. Without
    rotation.

    Parameters:
    d_l, d_b, d_r, d_t: [int or float] buffer distance at left, bottom, right
        and top sides.
    """

    def __init__(self, a=0, b=0, x=0, y=0, d_l=0, d_b=0, d_r=0, d_t=0, **kwargs):
        super(RectangleWithDistances, self).__init__(
            a=a, b=b, angle=0, x=x, y=y, **kwargs
        )

        self.distance_left = DistanceLeft(d_l, self)
        self.distance_bottom = DistanceBottom(d_b, self)
        self.distance_right = DistanceRight(d_r, self)
        self.distance_top = DistanceTop(d_t, self)

    @property
    def a_with_distances(self):
        return sum((self.distance_left.a, self.a, self.distance_right.a))

    @property
    def b_with_distances(self):
        return sum((self.distance_bottom.b, self.b, self.distance_top.b))

    @property
    def x_with_distances(self):
        """When in combination with y_with_distances: this point may intersect
        another shape.
        """
        return self.distance_left.x

    @property
    def y_with_distances(self):
        """When in combination with x_with_distances: this point may intersect
        another shape.
        """
        return self.distance_bottom.y

    @property
    def A_distances(self):
        """Return the area covered by distances."""
        return sum(
            (
                self.distance_left.A,
                self.distance_bottom.A,
                self.distance_right.A,
                self.distance_top.A,
            )
        )

    @property
    def A_with_distances(self):
        """Return the total area including distances."""
        return self.A + self.A_distances


class VisuDataLine(RectangleWithDistances, RectangleWithInner):
    """Representation of a Line area.

    text: [str] to draw in the center of the rectangle
    """

    conflict_category = 2

    def __init__(
        self,
        m=WIDTH_SAFE,
        n=LENGTH_SAFE,
        capacity=2,
        x=0,
        y=0,
        d_l=LINE_DISTANCE_A,
        d_b=LINE_DISTANCE_B,
        d_r=LINE_DISTANCE_A,
        d_t=LINE_DISTANCE_B,
        text="",
    ):
        super(VisuDataLine, self).__init__(
            m=m, n=n, count_inner=capacity, x=x, y=y, d_l=d_l, d_b=d_b, d_r=d_r, d_t=d_t
        )
        self.text = text

    @property
    def capacity(self):
        """Alias of self.count_inner."""
        return self.count_inner

    @property
    def util_rate_with_distances(self):
        return self.inner.A * self.count_inner / self.A_with_distances


class VisuDataDirectSingleRow(RectangleWithDistances, RectangleWithRotatableInner):
    """Representation of a single row direct area.

    angle_inner: makes sense until max. 75 degrees (with 12.5x3.5 slots)
    text: [str] to draw in the center of the rectangle
    """

    conflict_category = 3

    def __init__(
        self,
        m=LENGTH_SAFE,
        n=WIDTH_SAFE,
        angle_inner=45,
        capacity=1,
        x=0,
        y=0,
        d_l=DIRECT_DISTANCE_A,
        d_b=DIRECT_DISTANCE_B,
        d_r=0,
        d_t=DIRECT_DISTANCE_B,
        text="",
    ):
        super(VisuDataDirectSingleRow, self).__init__(
            m=m,
            n=n,
            count_inner=capacity,
            angle_inner=angle_inner,
            x=x,
            y=y,
            d_l=d_l,
            d_b=d_b,
            d_r=d_r,
            d_t=d_t,
        )
        self.text = text

    @property
    def capacity(self):
        """Alias of self.count_inner."""
        return self.count_inner

    @property
    def util_rate_with_distances(self):
        return self.inner.A * self.count_inner / self.A_with_distances


class VisuDataDirectSingleRow_90(RectangleWithDistances, RectangleWithRotatableInner):
    """Representation of a single row direct area.

    angle_inner: makes sense until max. 75 degrees (with 12.5x3.5 slots)
    text: [str] to draw in the center of the rectangle
    """

    conflict_category = 1

    def __init__(
        self,
        m=LENGTH_SAFE,
        n=WIDTH_SAFE,
        angle_inner=-45,
        capacity=1,
        x=0,
        y=0,
        d_l=0,
        d_b=DIRECT_DISTANCE_B,
        d_r=DIRECT_DISTANCE_A,
        d_t=DIRECT_DISTANCE_B,
        text="",
    ):
        super(VisuDataDirectSingleRow_90, self).__init__(
            m=m,
            n=n,
            count_inner=capacity,
            angle_inner=angle_inner,
            x=x,
            y=y,
            d_l=d_l,
            d_b=d_b,
            d_r=d_r,
            d_t=d_t,
        )
        self.text = text

    @property
    def capacity(self):
        """Alias of self.count_inner."""
        return self.count_inner

    @property
    def util_rate_with_distances(self):
        return self.inner.A * self.count_inner / self.A_with_distances


class VisuDataDirectDoubleRow(
    RectangleWithDistances, RectangleWithRotatedDoubleRowInner
):
    """Representation of a double row direct area.

    text: [str] to draw in the center of the rectangle
    """

    conflict_category = 4

    def __init__(
        self,
        m=LENGTH_SAFE,
        n=WIDTH_SAFE,
        capacity=1,
        x=0,
        y=0,
        d_l=DIRECT_DISTANCE_A,
        d_b=DIRECT_DISTANCE_B,
        d_r=DIRECT_DISTANCE_A,
        d_t=DIRECT_DISTANCE_B,
        text="",
    ):
        super(VisuDataDirectDoubleRow, self).__init__(
            m=m, n=n, count_inner=capacity, x=x, y=y, d_l=d_l, d_b=d_b, d_r=d_r, d_t=d_t
        )
        self.text = text

    @property
    def capacity(self):
        """Alias of self.count_inner."""
        return self.count_inner

    @property
    def util_rate_with_distances(self):
        return self.inner_left.A * self.count_inner / self.A_with_distances


def draw_rectangles(rectangles, draw_distances=True):
    """rectangles: [iterable] of Rectangle oder subclass objects."""
    fig = plt.gcf()
    ax = plt.gca()

    [p.remove() for p in reversed(ax.patches)]
    for r in rectangles:
        if isinstance(r, RectangleWithInner) or isinstance(
            r, RectangleWithRotatableInner
        ):
            ax.add_patch(
                patches.Rectangle(
                    (r.x, r.y), r.a, r.b, fill=r.fill, linestyle=r.linestyle
                )
            )
            # Inner rectangles
            for i in range(r.count_inner):
                ax.add_patch(
                    patches.Rectangle(
                        (r.x_inner(i), r.y_inner(i)),
                        r.inner.a,
                        r.inner.b,
                        r.inner.angle,
                        color=r.inner.color,
                        alpha=0.4,
                        fill=r.inner.fill,
                        linestyle=r.inner.linestyle,
                    )
                )

        if isinstance(r, RectangleWithRotatedDoubleRowInner):
            ax.add_patch(
                patches.Rectangle(
                    (r.x, r.y), r.a, r.b, fill=r.fill, linestyle=r.linestyle
                )
            )
            # Left inner rectangles
            for i in range(0, r.count_inner, 2):
                ax.add_patch(
                    patches.Rectangle(
                        (r.x_inner(i), r.y_inner(i)),
                        r.inner_left.a,
                        r.inner_left.b,
                        r.inner_left.angle,
                        color=r.inner_left.color,
                        alpha=0.4,
                        fill=r.inner_left.fill,
                        linestyle=r.inner_left.linestyle,
                    )
                )
            # Right inner rectangles
            for i in range(1, r.count_inner, 2):
                ax.add_patch(
                    patches.Rectangle(
                        (r.x_inner(i), r.y_inner(i)),
                        r.inner_right.a,
                        r.inner_right.b,
                        r.inner_right.angle,
                        color="red",
                        alpha=0.4,
                        fill=r.inner_right.fill,
                        linestyle=r.inner_right.linestyle,
                    )
                )

        # Vanilla case
        elif isinstance(r, Rectangle):
            ax.add_patch(
                patches.Rectangle(
                    (r.x, r.y),
                    r.a,
                    r.b,
                    r.angle,
                    color=r.color,
                    fill=r.fill,
                    alpha=r.alpha,
                    linestyle=r.linestyle,
                )
            )

        # Buffer distances
        if draw_distances and isinstance(r, RectangleWithDistances):
            for distance_rec in (
                r.distance_left,
                r.distance_bottom,
                r.distance_right,
                r.distance_top,
            ):
                if distance_rec.A > 0:
                    ax.add_patch(
                        patches.Rectangle(
                            (distance_rec.x, distance_rec.y),
                            distance_rec.a,
                            distance_rec.b,
                            fill=distance_rec.fill,
                            color=distance_rec.color,
                            alpha=distance_rec.alpha,
                            linestyle=distance_rec.linestyle,
                        )
                    )

        # Distances of a BinWithDistances
        if draw_distances and isinstance(r, BinWithDistances):
            for distance_rec in (
                r.distance_left_inner,
                r.distance_bottom_inner,
                r.distance_right_inner,
                r.distance_top_inner,
            ):
                if distance_rec.A > 0:
                    ax.add_patch(
                        patches.Rectangle(
                            (distance_rec.x, distance_rec.y),
                            distance_rec.a,
                            distance_rec.b,
                            fill=distance_rec.fill,
                            color=distance_rec.color,
                            alpha=distance_rec.alpha,
                            linestyle=distance_rec.linestyle,
                        )
                    )

        # Draw text if there is any
        if hasattr(r, "text") and r.text:
            box_props = dict(boxstyle="round", facecolor="white", alpha=0.5)
            ax.text(
                r.x_center, r.y_center, r.text, ha="center", va="center", bbox=box_props
            )

    plt.draw()


def draw_rectangles_newplot(
    rectangles, xlim=(0, 100), ylim=(0, 100), draw_distances=True
):
    """Initialize a new plot and draw rectangles.

    rectangles: sequence of Rectangle oder subclass objects
    xlim: [tuple] of x min and max
    ylim: [tuple] of y min and max
    """
    fig, ax = plt.subplots()
    draw_rectangles(rectangles, draw_distances)
    ax.set_aspect("equal")
    plt.xlim(*xlim)
    plt.ylim(*ylim)
    plt.show()
    return fig, ax


def packem(n=10):
    """Demo for packing without distances."""
    bin = Bin(150, 150)
    for ii in range(n):
        bin.items.append(VisuDataDirectSingleRow(capacity=randint(1, 5)))
        bin.items.append(VisuDataLine(capacity=randint(2, 4)))

    bin.pack()
    return bin


def packem_with_distances(n=5):
    """Demo for packing with distances."""
    bin = BinWithDistances(150, 150)
    for ii in range(n):
        bin.items.append(VisuDataDirectSingleRow(capacity=randint(1, 5)))
        bin.items.append(VisuDataLine(capacity=randint(2, 4)))
        bin.items.append(VisuDataDirectDoubleRow(capacity=randint(2, 5)))

    bin.pack()
    return bin


class Progressprinter:
    def __init__(self, until, interval=0.1):
        self.until = until
        self.last = 0
        self.interval = interval
        self.next = 0

    def notify(self, now):
        if self.next <= now:
            print("Progress: %d (%d %%)" % (now, now / self.until * 100))
            self.next += self.until * self.interval


def test_packing(tries=250, nmax=30):
    """Test for packing without distances."""
    propri = Progressprinter(tries)
    for i in range(tries):
        propri.notify(i)

        packem(randint(1, nmax))
    print("No error encountered in %d tries." % tries)


def test_packing_with_distances(tries=250, nmax=15):
    """Test for packing with distances."""
    propri = Progressprinter(tries)
    for i in range(tries):
        propri.notify(i)

        packem_with_distances(randint(1, nmax))
    print("No error encountered in %d tries." % tries)


lang_dict = {"Step": "Schritt", "av": "V", "no": "Nr."}


def tl(text):
    """Translate *text* using lang_dict."""
    if language == "en":
        return text
    else:
        return lang_dict[text]


if __name__ == "__main__":
    pass



"""Crawl spot prices and built data frame."""
import pandas as pd
import datetime
import pickle


class InputReader:
    LINK_INTRADAY = (
        "https://www.epexspot.com/en/market-data/intradaycontinuous/intraday-table/"
    )
    LINK_DAY_AHEAD = (
        "https://www.epexspot.com/en/market-data/dayaheadauction/auction-table/"
    )
    COUNTRY_INTRADAY = "/DE"
    COUNTRY_DAY_AHEAD = "/DE_LU/24"

    def __init__(self, type, start_date, end_date, safe_path):
        self.type = type  # "day_ahead" or "intraday"
        self.start_date = start_date
        self.end_date = end_date
        self.safe_path = safe_path

    def start_input(self):
        """Decides which input reader to run."""

        if self.type == "day_ahead":
            self.start_date += datetime.timedelta(days=6)
            self.input_day_ahead()
        elif self.type == "intraday":
            self.input_intraday()
        else:
            print("No valid type.")

    def input_day_ahead(self):
        """Input reader for https://www.epexspot.com/en/market-data/dayaheadauction/auction-table/2018-10-27/DE_LU/24"""

        date = self.start_date
        while date <= self.end_date:
            raw_pd_frame = pd.read_html(
                InputReader.LINK_DAY_AHEAD + str(date) + InputReader.COUNTRY_DAY_AHEAD
            )[8]
            pd_frame = raw_pd_frame.dropna()
            index = pd_frame.iloc[:, 0]

            i = 2
            while i < 9:
                values = pd_frame.iloc[:, i] / 1000
                series = pd.Series(list(values), index=index)
                date_name = str(
                    date + datetime.timedelta(days=(i - 8))
                )  # because of the +6 on top and the structure of the pd_frame it has to be +8
                pickle.dump(series, open(self.safe_path + date_name + ".p", "wb"))
                i += 1

            print(str(date) + ":imported")
            date = date + datetime.timedelta(days=7)

    def input_intraday(self):
        """Input reader for https://www.epexspot.com/en/market-data/intradaycontinuous/intraday-table/2018-10-18/DE"""
        date = self.start_date
        vector_row_needed = [
            2,
            3,
            5,
            6,
            9,
            10,
            12,
            13,
            16,
            17,
            19,
            20,
            23,
            24,
            26,
            27,
            30,
            31,
            33,
            34,
            37,
            38,
            40,
            41,
            44,
            45,
            47,
            48,
            51,
            52,
            54,
            55,
            58,
            59,
            61,
            62,
            65,
            66,
            68,
            69,
            72,
            73,
            75,
            76,
            79,
            80,
            82,
            83,
            86,
            87,
            89,
            90,
            93,
            94,
            96,
            97,
            100,
            101,
            103,
            104,
            107,
            108,
            110,
            111,
            114,
            115,
            117,
            118,
            121,
            122,
            124,
            125,
            128,
            129,
            131,
            132,
            135,
            136,
            138,
            139,
            142,
            143,
            145,
            146,
            149,
            150,
            152,
            153,
            156,
            157,
            159,
            160,
            163,
            164,
            166,
            167,
        ]
        while date < (self.end_date + datetime.timedelta(days=1)):
            date += datetime.timedelta(days=1)

            raw_pd_frame = pd.read_html(
                InputReader.LINK_INTRADAY + str(date) + InputReader.COUNTRY_INTRADAY
            )[0]
            values = raw_pd_frame.iloc[vector_row_needed, 7].astype("float") / 1000
            index = raw_pd_frame.iloc[vector_row_needed, 2]
            series = pd.Series(list(values), index=index)
            pickle.dump(
                series,
                open(
                    self.safe_path + str(date - datetime.timedelta(days=1)) + ".p", "wb"
                ),
            )
            print("Import complet: " + str(date - datetime.timedelta(days=1)))


class PowerFrame:
    """Constructs a pdFrame, which is used by the smart charging skript."""

    def __init__(self, max_sim_time, start_date, price_data_path):
        """Decides which pd constructer."""
        self.max_sim_time = max_sim_time
        self.start_date = start_date
        self.price_data_path = price_data_path

    def pdframe(self):
        """
        :return: power-pdframe with the power prices
        """
        power = pd.DataFrame(columns=["start_time", "end_time", "price"])
        current_day = self.start_date
        time_in_sec = 0  # sim time in sec
        while time_in_sec <= self.max_sim_time:
            power_price = pickle.load(
                open(self.price_data_path + str(current_day) + ".p", "rb")
            )
            time_delta = 86400 / power_price.size
            for price in power_price:
                row = [time_in_sec, time_in_sec + time_delta, price]
                power.loc[time_in_sec] = row
                time_in_sec += time_delta
            current_day += datetime.timedelta(days=1)

        return power



# -*- coding: utf-8 -*-
"""Load and check optimization and scenario settings.

EXEC_TIME: [str] top-level value appended during loading. Example:
      '2018-11-13_1800'

FILENAME: [str] top-level value appended during loading. filename argument
    that was passed to load_settings()
"""
import datetime
from eflips.helperFunctions import load_json


OPT_CONSTANTS = {"algorithm": {}, "scenario": {}}


def load_settings(filename):
    """Load custom opt settings from a json file.

    filename: [str] excluding file extension
    """
    OPT_CONSTANTS.clear()
    custom = load_json(filename)
    OPT_CONSTANTS.update(custom)

    # Append execution time
    OPT_CONSTANTS["EXEC_TIME"] = str(datetime.datetime.now().strftime("%Y-%m-%d_%H%M"))

    # Append filename
    OPT_CONSTANTS["FILENAME"] = filename
